/* $ZEL: sis1100_tmp_read.c,v 1.6.2.2 2003/08/07 11:51:43 wuestner Exp $ */

/*
 * Copyright (c) 2001-2003
 * 	Peter Wuestner.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/pci.h>

#include "sis1100_sc.h"

int
sis1100_tmp_read(struct SIS1100_softc* sc,
    	u_int32_t addr, int32_t am, u_int32_t size, int space,
        u_int32_t* data /* kernel virtual address, NOT user */
        )
{
    u_int32_t be, _data;
    u_int32_t error;
    u_int32_t head;

    be=((0x00f00000<<size)&0x0f000000)<<(addr&3);

    head=0x00000002|(space&0x3f)<<16|be;

    down(&sc->sem_hw);
    if (am>=0) {
        head|=0x800;
        sis1100writereg(sc, t_am, am);
    }
    sis1100writereg(sc, t_hdr, head);
    wmb();
    sis1100writereg(sc, t_adl, addr);
    mb();
    do {
	error=sis1100readreg(sc, prot_error);
    } while (error==0x005);
    rmb();
    _data=sis1100readreg(sc, tc_dal);
    up(&sc->sem_hw);

    switch (size) {
	case 4:
            *data=_data;
            break;
	case 2:
            *data=(_data>>((addr&2)<<3))&0xffff;
            break;
	case 1:
            *data=(_data>>((addr&3)<<3))&0xff;
	    break;
    }
    return error;
}

int
sis1100_remote_reg_read(struct SIS1100_softc* sc, u_int32_t offs,
    u_int32_t* data, int locked)
{
    u_int32_t error;

    if (!locked) down(&sc->sem_hw);
    *data=plxreadlocal0(sc, offs+0x800);
    rmb();
    error=sis1100readreg(sc, prot_error);
    if (!locked) up(&sc->sem_hw);
    return error;
}
